#!/usr/bin/python
# -*- coding: utf-8 -*-
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is SpellChecker version 1.0.
#
# The Initial Developer of the Original Code is European Environment
# Agency (EEA).  Portions created by Finsiel Romania are
# Copyright (C) European Environment Agency.  All
# Rights Reserved.
#
# Contributor(s):
# Soren Roug, EEA
#
#
#
# $Id: SpellChecker.py,v 1.10 2004/12/10 11:30:32 roug Exp $
#
import aspell
import string, sys
import urllib2, htmlentitydefs
from HTMLParser import HTMLParser
from cgi import parse_header
from AccessControl import ClassSecurityInfo
import Globals
from Globals import Persistent, Acquisition
import AccessControl
import OFS
from types import StringType


def parse_acceptlang(accept):
    """
    From a string formatted as specified in the RFC2616, it builds a data
    structure which provides a high level interface to implement language
    negotiation.
    returns the first element
    """
    x = accept.split(',')
    x = x[0].strip()
    l = x.split(';')
    return l[0]
 
class SpellChecker(
    OFS.SimpleItem.SimpleItem):
    "spell check webpages."

    # Declare the security
    #
    security=ClassSecurityInfo()


    manage_options=(
        {'label':'Properties', 'action':'manage_main'},
        {'label':'View', 'action':''},
        ) + OFS.SimpleItem.SimpleItem.manage_options

    meta_type = 'Spell Checker'

    security.declareProtected('View', 'index_html')
    index_html = Globals.DTMLFile("spellcheck", globals())

    security.declareProtected('View management screens', 'manage_main')
    manage_main = Globals.DTMLFile("edit_prop", globals())

    def __init__(self, id, title, http_proxy):
        self.id = id
        self.title = title
        self.http_proxy = http_proxy

    security.declareProtected('Change SpellCheckers', 'manage_edit')
    def manage_edit(self, title, http_proxy, REQUEST=None):
        """ Edits the checker's characteristics
        """
        self.title = title
        self.http_proxy = http_proxy

    ################################
    ##        Public methods       #
    ################################

    security.declareProtected('View', 'checkurl')
    def checkurl(self, url, languages=''):
        """ check url
            The languages argument is the language negotiation setting the user wants to
            send to the server. Defaults to the user's own browser setting.
        """
        url = string.strip(url)
        if url.lower()[:5] != "http:":
            raise IOError, "Only http is accepted"

        if self.http_proxy:
            _proxies = { 'http': self.http_proxy }
        else:
            _proxies = {}
        proxy_support = urllib2.ProxyHandler(_proxies)
        opener = urllib2.build_opener(proxy_support, urllib2.HTTPHandler)

        urllib2.install_opener(opener)

        req = urllib2.Request(url)
        if languages != '':
            req.add_header('Accept-Language', languages)
            pagelang = parse_acceptlang(languages)
        else:
            pagelang = 'en'
        conn = urllib2.urlopen(req)

        if not conn:
            raise IOError, "Failure in open"
        data = conn.read()
        headers = conn.info()
        conn.close()

        encoding = 'iso8859-1'  #Standard HTML
        if headers.has_key('content-type'):
            (ct, parms) = parse_header(headers['content-type'])
            if parms.has_key('charset'):
                encoding = parms['charset']

        lang_declared = 0
        if headers.has_key('content-language'):
            pagelang = headers['content-language']
            lang_declared = 1

        mhp = MyHTMLSpeller(pagelang, encoding)
        failure = ""
        try:
            mhp.feed(data)
        except:
            failure = sys.exc_info()[1]
        if mhp.foundlangdecl == 1:
            lang_declared = 1
        return { 'badwords':mhp.badwords, # Already in UTF-8
            'unknownlangs':mhp.unknownlangs,
            'encoding': mhp.encoding,
            'failure': failure,
            'language': mhp.lang,
            'lang_declared': lang_declared,
            'data': unicode(data, mhp.encoding).encode('utf-8').split('\n')
            }

    security.declareProtected('View', 'referer')
    def referer(self, REQUEST):
        """ Grab the HTTP_REFERER info and redirect to result page """
        if REQUEST.has_key('HTTP_REFERER'):
            REQUEST.RESPONSE.redirect(REQUEST['URL1'] + "?url=" + REQUEST['HTTP_REFERER'])
        else:
            REQUEST.RESPONSE.redirect(REQUEST['URL1'])


# Initialize the class in order the security assertions be taken into account
#
Globals.InitializeClass(SpellChecker)

def manage_addSpellChecker(self, id, title, http_proxy, REQUEST=None):
    """ Create a checker and install it in its parent Folder.
        The argument 'self' will be bound to the parent Folder.
    """
    ob = SpellChecker(id, title, http_proxy )
    self._setObject(id, ob)
    if REQUEST is not None:
        return self.manage_main(self, REQUEST)

manage_addSpellCheckerForm = Globals.DTMLFile('add_checker', globals())

transmap = {}
for char in string.punctuation + string.digits:
    transmap[ord(char)] = u" "
# Latin-1 punctuation
for char in range(160,192):
    transmap[char] = u" " 
# There's also general punctuation in the range 0x2000-0x206F
# But we ignore that (for now)

if transmap.has_key(39):
    del transmap[39]  # Single quote is used for contractions in english

# Elements we don't spellcheck as they contain programming language
ignore_elements = ( 'script', 'style' )

class MyHTMLSpeller(HTMLParser):

    def __init__(self, lang, encoding):
        HTMLParser.__init__(self)
        self.lang = lang
        self.foundlangdecl = 0
        self.encoding = encoding
        self.tagstack = []
        self.badwords = [] # The unknown words
        self.checked = {}  # The checked words
        self.checkflag = 1  # Are we in a tag we check?
        self.unknownlangs = []  # List of languages with no dictionary
        self.checkers  = {}  # spell checker connections
        self.__data = []

    def spellscan(self, str):
        """ Spellcheck a string
            String must be in UNICODE
        """
        if str == '':
            return
        if self.checkers.has_key(self.lang):
            speller = self.checkers[self.lang]
        else:
            # Setup a spell checker for the language
            # If not possible the checkers dict gets None for that language
            try:
                speller = aspell.new([ ('lang', self.lang),
                                       ('encoding', 'utf-8')])
            except:
                speller = None
            self.checkers[self.lang] = speller
        # Remove punctuation and numbers
        str = str.translate(transmap)
        # Split into words
        words = str.strip().split()
        if speller is None and self.lang not in self.unknownlangs:
            self.unknownlangs.append(self.lang)
        if speller is None or words == []:
            return
        badwords = []
        for word in words:
            if self.checked.has_key( (word, self.lang) ):
                continue   # Already checked
            self.checked[(word, self.lang)] = 1
            if speller.check(word.encode('utf-8')) == 0:
                badwords.append(word.encode('utf-8'))

        if badwords == []:
            return
        line = self.getpos()[0]
        for badword in badwords:
            suggestion = string.join(speller.suggest(badword),', ')
            self.badwords.append( (line, self.lang, badword, suggestion) )

    def meta_name(self, attrs):
        """ Look in meta tag for textual info"""
        foundit = 0
        # Is there a name attribute?
        for attr in attrs:
            if attr[0] == 'name' and string.lower(attr[1]) in ('description','keywords','title',
            'dc.description','dc.keywords','dc.title'
            ):
                foundit = 1
        if foundit == 0:
            return 0

        # Is there a content attribute?
        for attr in attrs:
            if attr[0] == 'content':
                data = attr[1]
                if type(data) == StringType:
                    data = unicode(data, self.encoding)
                self.spellscan(data)
        return 1

    def meta_encoding(self, attrs):
        """ Look in meta tag for page encoding (Content-Type)"""
        foundit = 0
        # Is there a content-type attribute?
        for attr in attrs:
            if attr[0] == 'http-equiv' and string.lower(attr[1]) == 'content-type':
                foundit = 1
        if foundit == 0:
            return 0

        # Is there a content attribute?
        for attr in attrs:
            if attr[0] == 'content':
                (ct, parms) = parse_header(attr[1])
                if parms.has_key('charset'):
                    self.encoding = parms['charset']
        return 1

    def meta_language(self, attrs):
        """ Look in meta tag for page language (Content-Language)
            Multiple languages MAY be listed for content that is intended for
            multiple audiences. The languages are comma-seperated.
            We just take the first one
        """
        foundit = 0
        # Is there a content-type attribute?
        for attr in attrs:
            if attr[0] == 'http-equiv' and attr[1].lower() == 'content-language':
                foundit = 1
        if foundit == 0:
            return 0

        # Is there a content attribute?
        for attr in attrs:
            if attr[0] == 'content':
                self.lang = parse_acceptlang(attr[1])
                self.foundlangdecl = 1
        return 1

    def xmllang_attr(self, attrs):
        "Check attributes for xml:lang or lang"
        for attr in attrs:
            if attr[0] in  ("lang", "xml:lang"):
                self.lang = attr[1]
                self.foundlangdecl = 1

    def handle_data(self, data):
        if self.checkflag == 1:
            if type(data) == StringType:
                data = unicode(data, self.encoding)
            self.__data.append(data)

    def handle_starttag(self, tag, attrs):
        if tag == 'meta':
            self.meta_encoding(attrs)
            self.meta_language(attrs)
            self.meta_name(attrs)

        self.xmllang_attr(attrs)
        self.tagstack.append((tag,self.lang))
        for attr in attrs:
        #sgmllib doesn't handle entities :-(
            if attr[0] in  ("title", "alt", "abbr", "summary"):
                data = attr[1]
                if type(data) == StringType:
                    data = unicode(data, self.encoding)
                self.spellscan(data)
        if tag in ignore_elements:
            self.checkflag = 0

    def handle_endtag(self, tag):
        """ Spell check the data coming before the tag
        """
        self.spellscan(string.join(self.__data,''))
        self.__data = []
        self.tagstack.pop()
        if len(self.tagstack) != 0:
            self.lang = self.tagstack[0][1]
        if tag in ignore_elements:
            self.checkflag = 1

    def handle_charref(self, name):
        """ Handle character reference for UNICODE
        """
        if name[0] in ('x', 'X'):
            try:
                n = int(name[1:],16)
            except ValueError:
                return
        else:
            try:
                n = int(name)
            except ValueError:
                return
        if not 0 <= n <= 65535:
            return
        self.handle_data(unichr(n))

    def handle_entityref(self, name):
        """Handle entity references.
        """
        table = htmlentitydefs.name2codepoint
        if name in table:
            self.handle_data(unichr(table[name]))
        else:
            return
